// An image (= matrix) class, with a few tools
// Simple loading-from-file and display (text mode)
// For more practical loading/saving (from/to png, jpg, ... files) or display (color pixels in a window), see with_graphical_library.h, which is based on an external graphical library

#ifndef IMAGES_H
#define IMAGES_H

#include <cmath>
#include <unistd.h>
#include <algorithm>
#include <cassert>

#include <fstream>
#include <string>
#include <iostream>
using namespace std;


namespace MRF_F0 {

// A class Image, to contain a 2D-table of int, double or anything else.

template <class precis>
struct RecImage {

  int nx, ny;
  precis * val;

  // constructor

  RecImage(){
    nx = 0; ny = 0;
  }

  // constructor allocating the size required

  RecImage(int x, int y){
    allocate(x,y);
  }

  // constructor copying a pointer to the table
  RecImage(precis* tab, int x, int y){
    nx = x; ny = y; val = tab;
  }

  // constructor initializing with an image from a file, and applying a zoom factor

  RecImage(const char * fich, double zoom = 1);
  void load(const char* fich, double zoom = 1);
  void load(const string fich, double zoom = 1){
    load(fich.c_str(), zoom);
  }

  // manual desctructor ... not automatic !

  void empty(){
    if ((nx!=0)&&(ny!=0)){
      nx = 0; ny = 0;
      delete[] val;
    }
  }  

  /*
  // why not ? because of pointers shared for certain types of copies.
  ~RecImage(){ empty(); }
  */

  // To allocate by hand

  void allocate(int x, int y){
    nx = x; ny = y;
    if ((x!=0)&&(y!=0))
      val = new precis[nx*ny]; 
  }

  // To handle the table easily and carfully
  
  inline precis& operator()(int x, int y){
    assert((x<nx)&&(y<ny)&&(x>=0)&&(y>=0));
    return val[x+nx*y];
  }

  inline precis operator()(int x, int y) const {
    assert((x<nx)&&(y<ny)&&(x>=0)&&(y>=0));
    return val[x+nx*y];
  }

  void operator+=(const RecImage<precis>& B){
    assert(nx == B.nx); assert(ny == B.ny);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	(*this)(i,j) += B(i,j);
  }

  void operator-=(const RecImage<precis>& B){
    assert(nx == B.nx); assert(ny == B.ny);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	(*this)(i,j) -= B(i,j);
  }

  RecImage<precis> operator+(const RecImage<precis>& B) const {
    assert(nx == B.nx); assert(ny == B.ny);
    RecImage<precis> C(nx,ny);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	C(i,j) = (*this)(i,j) + B(i,j);
    return C;
  }

  RecImage<precis> operator-(const RecImage<precis>& B) const {
    assert(nx == B.nx); assert(ny == B.ny);
    RecImage<precis> C(nx,ny);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	C(i,j) = (*this)(i,j) - B(i,j);
    return C;
  }

  // To have test images
  void fill(precis p); // fill with value p
  void nul();  // all coefficients to value 0

  // To read an image from a simple text format file (see code)
  void read_image_from_textfile(const char * fich);

  // Display with possibly a zoom factor and a translation

  void display(double zoom = 1, int p = 0, int q = 0);
  void save(const char * fich, double zoom = 1);
  void save(string fich, double zoom = 1){
    save(fich.c_str(), zoom);
  }
  void display_text(int prec = -1) const;

};


// Common types

typedef RecImage<int> IntImage;
typedef RecImage<double> DoubleImage;
//typedef RecImage<float> FloatImage;
//typedef RecImage<float> DoubleImage;
typedef RecImage<bool> BoolImage;


// A few declarations

template <class precis1, class precis2>
void copy(const RecImage<precis1>& A, RecImage<precis2>& B);


// Basic maths

inline int square(int a){ return (a*a);}
inline double square(double a){ return (a*a);}

// Useful tool

inline int limite(int a, int b, int c){
  if (b<a) return a;
  if (b>c) return c;
  return b;
}

inline int limite(int a, double b, int c){
  return limite(a,int(b),c);
}



// Read an image from a file in an homemade text format
// Put the full path + name of file in 'fich'
// To read a png, jpg (etc.) image from a file, you should use with_graphical_library.h, which makes use of an external graphical library
template <class precis> 
void RecImage<precis>::read_image_from_textfile(const char * fich) {
  //  string fichier = "/users/odyssee/charpiat/recalage/images/" + string(fich) + ".dat";
  string fichier(fich);
  ifstream f;
  f.open(fichier.c_str());
  f >> nx;
  f >> ny;
  cout << "Size of the image loaded: " << nx << " x " << ny << endl;
  allocate(nx, ny);
  for (int i=0; i<nx; i++)
    for (int j = 0; j<ny; j++){
      f >> (*this)(i,j);
      //      cout << U(i,j);
    }

}



// Displays the matrix in text mode in the terminal
// prec = nb of digits
// To display images in a window (as pixels), you should use with_graphical_library.h, which makes use of an external graphical library
template <class T>
void RecImage<T>::display_text(int prec) const{
  cout << endl << "Size: " << nx << " x " << ny << endl;
  for (int i = 0; i<nx; i++){
    int nbd =  int( ceil( log(nx)/log(10) ) );
    char* b = new char[nbd+1];
    sprintf(b, "%*d", nbd, i);
    cout << "L " << string(b) << " :  ";
    delete[] b;
    
    for (int j = 0; j<ny; j++){
      if (prec == -1){ // standard case
	cout << "  " << (*this)(i,j);
      } else { // case with alignment
	char* c = new char[prec+5];
	sprintf(c, "% -*.*g", prec+2, prec, double ( (*this)(i,j) ) );
	cout << "  	" << string(c);
	delete[] c;
      }
    }
    cout << endl;
  }
}






// Fill with 0
template <class precis>
void RecImage<precis>::nul(){
  for (int i = 0; i < nx; i++)
    for (int j = 0; j < ny; j++)
      (*this)(i,j) = precis(0);
}


// Multiply an image by a constant factor

template <class precis1>
void multiply(RecImage<precis1>& A, double z){
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      A(i,j) *= z;
}


// Copies element by element a matrix into a new one.
template <class precis1, class precis2>
void copy(const RecImage<precis1>& A, RecImage<precis2>& B){
  if ((B.nx != A.nx)||(B.ny != A.ny)){
    B.empty();
    B.allocate(A.nx, A.ny);
  }

  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      B(i,j) = precis2(A(i,j));
}

// Copies and deletes the original one
template <class precis>
void copy_suppr(RecImage<precis>& A, RecImage<precis>& B){
  B.empty();
  B.val = A.val;
  B.nx = A.nx;
  B.ny = A.ny;
  A.nx = 0;
  A.ny = 0;
  A.val = NULL;
}


// Max norm (infinity norm) of a matrix
double norm_max(const DoubleImage& A){
  double z = 0;
  double nor;
  for (int i =0; i<A.nx; i++)
    for (int j =0; j<A.ny; j++){
      nor = fabs( A(i,j) );
      if (nor>z) z = nor;
    }
  return z;
}

// Smallest coefficient in the matrix
double smallest_coeff(const DoubleImage& A){
  double z = A(0,0);

  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      if (A(i,j) < z) z = A(i,j);

  return z;
}



// Smooth an image/matrix with a Gaussian blur 
// Case of matrix of integers
void smooth(IntImage& A, double sigma){
  DoubleImage B;
  copy(A, B);
  
  double gamma = 1/(2*sigma*sigma);
  int larg = min(2*max(A.nx, A.ny), 2*max(1,int(sigma))+1);

  int zero = larg/2;
  int tx = A.nx; int ty = A.ny;
  
  DoubleImage masque(larg, larg);
  
  for (int i =0; i<larg; i++)
    for (int j = 0; j < larg; j++)
      masque(i,j) = exp(-gamma*sqrt(double(square(i-zero)+square(j-zero))));
  
  for (int i=0; i<tx; i++)
    for (int j =0; j<ty; j++){
      
      int kmin = limite(0,i-zero,tx-1);
      int kmax = limite(0,i+zero,tx-1);
      int lmin = limite(0,j-zero,ty-1);
      int lmax = limite(0,j+zero,ty-1);
      
      B(i,j) = 0;
      double poids = 0;
      for (int k = kmin; k<=kmax; k++)
	for(int l = lmin; l<=lmax; l++){
	  poids += masque(k-i+zero, l-j+zero);
	  B(i,j) += masque(k-i+zero,l-j+zero)*A(k,l);
	} 
      B(i,j) /= poids;
    }  
  
  copy(B, A);
  B.empty();
  masque.empty();
}

// Case of matrix of doubles
void smooth(DoubleImage& A, double sigma, double coeffsigma = 3){
  DoubleImage B;
  copy(A, B);
  
  double gamma = 1/(2*sigma*sigma);
  int larg = min(2*max(A.nx, A.ny), int(2*coeffsigma*max(1,int(sigma))+1));

  int zero = larg/2;
  int tx = A.nx; int ty = A.ny;
  
  DoubleImage masque(larg, larg);
  
  for (int i =0; i<larg; i++)
    for (int j = 0; j < larg; j++)
      masque(i,j) = exp(-gamma*sqrt(double(square(i-zero)+square(j-zero))));
  
  for (int i=0; i<tx; i++)
    for (int j =0; j<ty; j++){
      
      int kmin = limite(0,i-zero,tx-1);
      int kmax = limite(0,i+zero,tx-1);
      int lmin = limite(0,j-zero,ty-1);
      int lmax = limite(0,j+zero,ty-1);
      
      B(i,j) = 0;
      double poids = 0;
      for (int k = kmin; k<=kmax; k++)
	for(int l = lmin; l<=lmax; l++){
	  poids += masque(k-i+zero, l-j+zero);
	  B(i,j) += masque(k-i+zero,l-j+zero)*A(k,l);
	} 
      B(i,j) /= poids;
    }  
  
  copy(B, A);
  B.empty();
  masque.empty();
}


// Copies with a zoom factor (a bit rough)
template <class precis, class precis2>
void copy_fact(RecImage<precis>& A, RecImage<precis2>& B, double zoom){

  B.empty();
  int tx = int(A.nx * zoom); 
  int ty = int(A.ny * zoom);
  B.allocate(tx, ty);
  
  double zx = double(A.nx)/B.nx;
  double zy = double(A.ny)/B.ny;

  for (int i=0; i<B.nx; i++)
    for (int j = 0; j<B.ny; j++)
      B(i,j) = precis2(A(limite(0,zx*i,A.nx-1), limite(0,zy*j,A.ny-1)));
}


// Fill with a constant value (supposes allocation has already been done)
template <class precis>
void RecImage<precis>::fill(precis p){
  for (int i = 0; i<nx; i++)
    for (int j = 0; j<ny; j++)
      (*this)(i,j) = p;
}


// Infinite norm between 2 matrices
// T : int, double, float... or else add .norme()
template <class T>
double difference_max(const RecImage<T>& A, const RecImage<T>& B){
  double x = 0;
  assert(A.nx == B.nx);
  assert(A.ny == B.ny);
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++){
      double dx = fabs(A(i,j) - B(i,j));
      if (dx > x)
	x = dx;
    }
  return x;
}


} // end of namespace
#endif
