// Minimization of energies in F0, with graph cuts, once they are identified

#ifndef MINIMIZATION_OF_MRF_ONCE_IDENTIFIED
#define MINIMIZATION_OF_MRF_ONCE_IDENTIFIED


// MRF - graph-cut by Kolmogorov, Veksler et al
#include "mrf.h"
#include "ICM.h"
#include "GCoptimization.h"
#include "MaxProdBP.h"
#include "graph.h"

// F0 library
#include "familial_permutations.h"

// Random generator
#include "randomness.h"



namespace MRF_F0 {

// Global variables for display         
// not used, actually
// bool verbose = false;

// Let's reach infinity !
//double infini = 100000000;


// A few simple tools

void error_display(char* x){
  cout << endl << "Error message from graph cut : " << endl
       << x << endl;
}





/*************************** Minimization of MRF recognized as being in F0 *********************/


// Minimize a given MRF in F0 (given with alphas, permutations, multiple possible interpretations and choice of interpretation).
// Return the solution (one label per variable)
vector<int> minimize_familial_MRF(const Energy& E, const PreGraphMulti& PG, bool test_local = true, bool verbose_base = false, bool verbose_graph = false){

  /********* let's build the graph *******/
  
  /* // How it works in the binary graph cut library by Kolmogorov & al
     void Graph::add_edge(node_id from, node_id to, captype cap, captype rev_cap)
     void Graph::set_tweights(node_id i, captype cap_source, captype cap_sink)

     void* n1 = G.add_node();
     void* n2 = G.add_node();
     G.add_edge(n1, n2, 3, 3);
     G.set_tweights(n1, 7, 1);
     G.set_tweights(n2, 7, 1);
     
     float energy = G.maxflow();
     
     int a = G.what_segment(n1); // 0 = source, 1 = sink
  */
  
  
  //Building nodes and links to source/sink

  Graph G(error_display);
  
  typedef void* Nnode; // that's a node in a graph
  typedef vector<Nnode> Variable; // when you need several nodes to represent one variable
  // Beware the lonely variables
  
  const vector<int>& ChoixInterp = PG.Solutions.front(); 
  
  vector< Variable > Rep(E.nb_var);
  vector<int> taille_var(E.nb_var);
  double poids_tot = 0;
  for (int i = 0; i<E.nb_var; i++){
    taille_var[i] = int( log( E.Pot[i].size() ) / log(2) + 0.1);
    if (! PG.solitary[i]){ // if x_i is lonely, don't minimize
      const vector<double>& Coeff = PG.Coeff_Pot[i][ ChoixInterp[i] ];
      
      for (int k = 0; k<taille_var[i]; k++){
	Rep[i].push_back( G.add_node() );
	
	double poids = Coeff[k];
	if (poids > 0)
	  G.set_tweights( Rep[i].back(), poids, 0 ); // NB : energy is multiplied by 2 in order not to be uselessly disturbed
	else
	  G.set_tweights( Rep[i].back(), 0, -poids );
	poids_tot += fabs(poids);
      }    
      
      // Adding the intra_nodes
      // They are the complementary indices of Coeff_Pot
      int c = taille_var[i];
      for (int k1 = 0; k1<taille_var[i]; k1++)
	for (int k2 = k1+1; k2<taille_var[i]; k2++){
	  G.add_edge( Rep[i][k1], Rep[i][k2], Coeff[c], Coeff[c] );
	  c++;
	}
    }
  }
  
  for (int i = 0; i<E.nb_var; i++)
    for (int j = 0; j< int( E.Neighbors[i].size() ); j++){
      int vj = E.Neighbors[i][j].first;
      if (i<vj){ // doing everything twice is useless
	int num_rep_ij = PG.Coherence[i][ ChoixInterp[i] ].second[j].first; // : correct version of M_ij in AlpesL[...][...]
	int num_mat = E.Neighbors[i][j].second;
	DoubleImage& A = * PG.AlpesL[num_mat][num_rep_ij].alpha;
	const vector<int>& Perm_node_i = PG.Coherence[i][ ChoixInterp[i] ].second[j].second; // permut' to apply on side x_i
	
	int iv = find_value_1(E.Neighbors[vj], i); // i is the iv-th neighbor of vj, as well as vj is the j-th neighbor of i
	const vector<int>& Perm_node_j = PG.Coherence[vj][ ChoixInterp[vj] ].second[iv].second; // permut' to apply to side x_j
	
	for (int ki = 0; ki < taille_var[i]; ki++)
	  for (int kj = 0; kj < taille_var[vj]; kj++){ /*******************************************************************/
	    double poids =  A( Perm_node_i[ki], Perm_node_j[kj] ); // no multiplication by 2 !???
	    //	    cout << "p : " << poids << endl;
	    G.add_edge( Rep[i][ki], Rep[vj][kj], poids, poids);
	  }	
      }
    }
  
  //  cout << "Size of float : " << sizeof(float) << endl;
  //  cout << "Size of double : " << sizeof(double) << endl;



  /****** We apply the graph cut algorithm ******/

  float alors = 2 * G.maxflow();
  if (verbose_base)
    cout << "Energy at minimum (graph cut) : " << alors << endl;

  // Interprétation...
  
  vector<int> Resultat(E.nb_var);
  
  for (int i = 0; i<E.nb_var; i++){
    if (PG.solitary[i]){ // solitary variable ? special treatment
      Resultat[i] = find_smallest(E.Pot[i]); // label already real, not binary in this case
    } else {
      
      // Binary interp. of x_i : 001 --> 4 :  Coherence[i][ ChoixInterp[i] ].first
      // no suplpementary permut' over potentials (pffew!)
      
      int labl_bin = 0;
      for (int k = 0; k<taille_var[i]; k++)
	labl_bin += G.what_segment(Rep[i][k]) * int( pow(2,k) + 0.1 ); // 0 = source, 1 = sink    
      
      Resultat[i] = PG.Coherence[i][ ChoixInterp[i] ].first[labl_bin];
    }
  }

  //  alors += E.constante - poids_tot;
  //  if (verbose_base)
  //    cout << "Energy at minimum (with constant) : " << alors << endl;


  double energy_min = evaluate_energy(E, Resultat);
  if (verbose_graph)
    cout << "Cost of this labelling choice : " << energy_min << endl;


  
  /**** We display the solution ****/
  
  if (verbose_graph){

    cout << "Result : " << Resultat << endl;
    
    for (int i = 0; i<E.nb_var; i++)
      cout << "Pot " << i << " : " << E.Pot[i] << endl;
    
    for (int i = 0; i<E.nb_var; i++)
      for (int j = 0; j<int(E.Neighbors[i].size()); j++)
	if (i< E.Neighbors[i][j].first ) 
	  cout << "M_(" << i << "," << E.Neighbors[i][j].first << ") : " << E.Interact[ E.Neighbors[i][j].second ].first << endl;
  }
    
  /**** We test ****/

  if (test_local){
    bool cava = true;
    for (int i = 0; i<E.nb_var; i++){
      int v = Resultat[i];
      for (int k = 0; k<int(E.Pot[i].size()); k++)
	if (k != v){
	  Resultat[i] = k;
	  double r = evaluate_energy(E, Resultat);
	  if (r < energy_min){
	    cava = false;
	    if (verbose_graph)
	      cout << "Smaller : " << Resultat << " : " << r << endl;
	  }
	}
      Resultat[i] = v;
    }

    if (cava){
      if (verbose_base)
	std::cout << TextGreen << "+" << TextNormal; std::cout.flush();
    } else {
      if (verbose_graph)
	cout << BoldRed << "*****************************************************************************************************************" << TextNormal << endl;
      else {
	std::cout << BoldRed << "*" << TextNormal; std::cout.flush();
      }
    }
  }
  
  return Resultat;
}







// The same as above but with a simpler representation of the energy (case where the interpretation is given directly, instead of a list of interepretations and an index to pick one)
// Prend une MRF de la famille reconnue (avec son écriture alpha, interprétation unique) et la minimise.
// Renvoie la solution (une étiquette par variable)
vector<int> minimize_familial_MRF(const PreGraph& PG, const Energy& E,  bool test_stupide = true, bool test_local = false, bool test_global = false, bool verbose_base = false, bool verbose_graph = false, bool verbose_a_lot = false){

  /********* on génère le graph *******/
  
  //création des nodes et lien des étiquettes extrêmes aux sources/puits
  Graph G(error_display);
  
  int nb_var = E.nb_var; //int(PG.Coeff_Pot.size());
  typedef void* Nnode; // ben, un node du graph, quoi
  typedef vector<Nnode> Variable; // il faut plusieurs nodes pour représentater une variable
  vector< Variable > Rep(nb_var);

  vector<int> taille_var(nb_var); //nb nodes pour chaque variable
  double poids_tot = 0;
  for (int i = 0; i<nb_var; i++){
    if (! PG.solitary[i]){ // si x_i est solitaire, on ne minimise pas
      taille_var[i] = int( log(E.Pot[i].size()) / log(2) + 0.1 ) ; //int( PG.Coeff_Pot[i].size() ); 
      assert(taille_var[i] != 0);
      
      for (int k = 0; k<taille_var[i]; k++){
	Rep[i].push_back( G.add_node() );
	
	double poids = PG.Coeff_Pot[i][k];
	//	cout << poids << "  " ;
	if (poids > 0)
	  G.set_tweights( Rep[i].back(), poids, 0 ); // NB : on multiplie l'énergie par 2 pour ne pas s'membêter inutilement.
	else
	  G.set_tweights( Rep[i].back(), 0, -poids );
	poids_tot += fabs(poids);
      }    
      
      // Ajout des intra_nodes
      // ce sont les numéros complémentaires de Coeff_Pot
      int c = taille_var[i];
      for (int k1 = 0; k1<taille_var[i]; k1++)
	for (int k2 = k1+1; k2<taille_var[i]; k2++){
	  G.add_edge( Rep[i][k1], Rep[i][k2], PG.Coeff_Pot[i][c], PG.Coeff_Pot[i][c] );
	  //	  cout << PG.Coeff_Pot[i][c] << "   ";
	  c++;
	}
    }
  }
  

  for (int m = 0; m<int(E.Interact.size()); m++){

    const DoubleImage& A = PG.Alpha[m];
    int i = E.Interact[m].second.first;
    int j = E.Interact[m].second.second;

    for (int ki = 0; ki < taille_var[i]; ki++)
      for (int kj = 0; kj < taille_var[j]; kj++){ /*******************************************************************/
	double poids =  A(ki,kj);                 // BEUH ? FAUT PAS MULTIPLIER PAR 2 ???????
	G.add_edge( Rep[i][ki], Rep[j][kj], poids, poids);
	//	cout << poids << "   ";
      }	
  }

  //  cout << endl;
  /****** on applique l'algo de graph cut ******/

  float alors = G.maxflow();
  if (verbose_base)
    cout << "Energy at minimum (graph cut) : " << alors << endl;

  // Interprétation...
  
  vector<int> Resultat(nb_var);
  
  for (int i = 0; i<nb_var; i++){
    if (PG.solitary[i]){ // les variables solitary, c'est différent
      Resultat[i] = find_smallest(E.Pot[i]); // étiquette déjà réelle, pas binaire dans ce cas
    } else {
      
      // Interp binaire x_i : 001 --> 4 :  Coherence[i][ ChoixInterp[i] ].first
      // pas de permut' supplémentaires sur les potentials (ouf!)
      
      int labl_bin = 0;
      for (int k = 0; k<taille_var[i]; k++)
	labl_bin += G.what_segment(Rep[i][k]) * int( pow(2,k) + 0.1); // 0 = source, 1 = sink    
      
      Resultat[i] = PG.Interp[i][labl_bin];
    }
  }




  
  /**** on display la solution ****/
  
  if (verbose_a_lot){

    cout << "Result : " << Resultat << endl;
    

    cout << endl << "Initial energy :" << endl;
    for (int i = 0; i<E.nb_var; i++)
      cout << "Pot " << i << " : " << E.Pot[i] << endl;
    
    for (int m = 0; m<int(E.Interact.size()); m++)
      cout << "M_(" << E.Interact[m].second.first << "," << E.Interact[m].second.second << ") : " << E.Interact[m].first << endl;


    cout << endl << endl << "Approximated energy :" << endl;
    for (int i = 0; i<E.nb_var; i++){
      if (PG.solitary[i])
	cout << "(solitary) Pot " << i << " : " << E.Pot[i] << endl;
      else {
	vector<double> V = alphas_and_interp_to_potential(PG.Coeff_Pot[i], PG.Interp[i]);
	cout << "Pot " << i << " : " << V << endl;
	V.clear();
      }
    }    

    for (int m = 0; m<int(E.Interact.size()); m++){ // alpha_to_matrix o interp
      DoubleImage A = alphas_and_interp_to_matrix( PG.Alpha[m], PG.Interp[ E.Interact[m].second.first ], PG.Interp[ E.Interact[m].second.second ] );
      cout << "M_(" << E.Interact[m].second.first << "," << E.Interact[m].second.second << ") : " << A << endl;
      A.empty();
    }


  }
    

  double energy_min = evaluate_energy(E, Resultat);
  if (verbose_graph){
    cout << endl << "Result : " << Resultat << endl; // (again)
    cout << "Cost of this labelling choice : " << energy_min  << endl;
  }

  /**** on teste ****/


  // test stupide
  //  bool test_stupide = true;
  double r_stupide;
  if (test_stupide){
    vector<int> Coord(nb_var);
    for (int i = 0; i<nb_var; i++)
      Coord[i] = find_smallest(E.Pot[i]); // étiquette déjà réelle, pas binaire dans ce cas
    r_stupide = evaluate_energy(E, Coord);
    cout << endl << "Test of (very-)locally optimal variables : " << r_stupide << endl;
  }

  double r_local;
  if (test_local){
    cout << endl << "Local test (changing one variable each time) :" << endl;
    bool cava = true;
    int plus_grands = 0;
    double vpg = 0;
    int plus_petits = 0;
    double vpp = 0;
    double vmin = energy_min;
    for (int i = 0; i<E.nb_var; i++){
      int v = Resultat[i];
      for (int k = 0; k<int(E.Pot[i].size()); k++)
	if (k != v){
	  Resultat[i] = k;
	  double r = evaluate_energy(E, Resultat);
	  if (r < energy_min){
	    cava = false;
	    if (verbose_a_lot)
	      cout << "Smaller : " << Resultat << " : " << r << endl;
	    plus_petits ++;
	    vpp += r;
	    if (r<vmin)
	      vmin = r;
	  } else {
	    plus_grands ++;
	    vpg += r;
	  }
	}
      Resultat[i] = v;
    }

    if (plus_petits != 0) vpp /= plus_petits;
    cout << "Smaller ones: " <<  plus_petits << " cases "; if (plus_petits != 0) cout << ": " << vpp << " (average)"; cout << " ; minimum found : " << vmin << endl;
    if (plus_grands != 0) vpg /= plus_grands;
    cout << "Bigger ones: " <<  plus_grands << " cases "; if (plus_grands != 0) cout << ": " << vpg << " (average)"; cout << "." << endl;

    if (verbose_graph)
      cout << "Cost os this labelling choice : " << BoldMauve << energy_min << TextNormal << endl;

    if (cava){
      if (verbose_base)
	std::cout << TextGreen << "+" << TextNormal; std::cout.flush();
    } else {
      if (verbose_graph)
	cout << BoldRed << "*****************************************************************************************************************" << TextNormal << endl;
      else {
	std::cout << BoldRed << "*" << TextNormal; std::cout.flush();
      }
    }
    r_local = vmin;
  }




  //  bool test_global = true;
  double r_global;
  if (test_global){
    int nb_tests = 10000;
    cout << endl << "Random test, " << nb_tests << " tries." << endl;
    int plus_grands = 0;
    double vpg = 0;
    double vmin = energy_min;
    double sigma_g = 0;
    int plus_petits = 0;
    double vpp = 0;
    vector<int> Coord(E.nb_var);
    for (int k = 0; k<nb_tests; k++){
      for (int i = 0; i<E.nb_var; i++)
	if (PG.solitary[i])
	  Coord[i] = Resultat[i];
	else 
	  Coord[i] = rand() % int(E.Pot[i].size());
      
      double r = evaluate_energy(E, Coord);
      if (r < energy_min){
	//	    if (verbose_graph)
	//	      cout << "Plus petit : " << Resultat << " : " << r << endl;
	plus_petits++;
	vpp += r;

	if (r<vmin)
	  vmin = r;
      } else {
	plus_grands ++;
	vpg += r;
	sigma_g += r*r;
      }

    }

    if (plus_petits != 0) vpp /= plus_petits;
    cout << "Smaller ones: " <<  plus_petits * 1. / nb_tests << " of cases "; if (plus_petits != 0) cout <<  ": " << vpp << " (average)"; cout << "." << endl;
   
    if (plus_grands != 0) { vpg /= plus_grands; sigma_g /= plus_grands; sigma_g -= vpg*vpg; sigma_g = pow(sigma_g, 0.5); }
    cout << "Bigger grands: " <<  plus_grands * 1. / nb_tests << " of cases : " << vpg << " in average, with standard deviation " << sigma_g << endl;

    r_global = vmin;
    Coord.clear();
  }
  

  if (verbose_base){
    cout << endl;
    if (test_stupide || test_local || test_global)
      cout << "Summary of tests : " << endl;
    if (test_stupide){
      if (r_stupide < energy_min)
	cout << BoldRed;
      else
	cout << BoldGreen;
      cout << "Stupide";
    }
    
    if (test_local){
      if (r_local < energy_min)
	cout << BoldRed;
      else
	cout << BoldGreen;
      cout << "  Local ";
    }

    if (test_global){
      if (r_global < energy_min)
	cout << BoldRed;
      else
	cout << BoldGreen;
      cout << " Global";
    }

    cout << TextNormal << endl;
  }


  return Resultat;
}







/************************* Random test (random energy in F0, then recognition, and minimization) ******************/

// You might get inspiration from that code to minimize your own energy
// Lines to update are in the first sections, and marked with <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
// They are also regrouped in a first /***commented***/ section.

// Generates a random energy in F0, then recognizes it and decomposes it within the family, then minimization with a graph cut
// (minimization done with the functions above)
// Booleans : level of details in the algorithm outputs
// Verbose_basic : basic information (names of steps)
// Verbose_summaries : display also important information at each step (e.g., solutions found)
// Verbose_a_lot : display potentials, permutations and matrices of the initial problem (before/after permutations)
// Verbose_coherence : details, variable per variable, of the search for locally-possible interpretations
// Verbose_coherence_pot : idem concerning the condition that potentials should be admissible for these interpretations
// Verbose_global : details during exploration (search for coherent interpretations)
// Test_canonisation : check the canonization did not change the energy
// Verbose_multi : details when multiple interpretations are possible
// Verbose_graph : details about the graph

 void minimize_random_familial_energy(bool verbose_basic = true, bool verbose_summaries = true, bool verbose_a_lot = false, bool verbose_coherence = false, bool verbose_coherence_pot = false, bool verbose_global = false, bool test_canonisation = false,  bool verbose_multi = false, bool verbose_graph = false, ostream& cout = std::cout){


  if (verbose_basic){
    cout << TextGreen << "**********************************************************" << endl;
    cout << "*****************" << TextRed << "  Creating an energy  " << TextGreen << "*******************" << endl;
    cout << "**********************************************************" << TextNormal << endl;
  }

  /**************************************************************/
  /************** Code to create you own energy... **************/
  /**************************************************************/

  /*****
  int nb_var = ...; // Number of variables 

  vector< int > NbLabl(nb_var); // nb possible labels per variable x_i   // not very useful, since it will be  Pot[i].size()
  for (vector<int>::iterator i = NbLabl.begin(); i != NbLabl.end(); ++i)  
    (*i) = ... ; // number of possible labels for variable x_i

  vector< vector<double> > Pot(nb_var); // potentials of each label for each variable
  for (int i = 0; i<nb_var; i++){
    Pot[i] = ... ;  // Put here the individual potential of variable x_i, as a vector<double>
  }

  vector< vector< pair<int, int> > > Neighbors(nb_var); // list, for each variable, of the indices of its neighbors (with the index of the associated interaction matrix), initialized with "no neighbor"
  //  typedef vector< pair<DoubleImage, pair<int,int> > >   MIV; // to remember the type used in what follows
  Energy::MIV Interact; // Interaction matrices with the numberings of the variables involved
  int nb_mat_inter = 0;

  for (int i = 0; i<nb_var; i++)
    for (int j = i+1; j<nb_var; j++)
      if ( ... variables x_i and x_j are neighbors && (i<j)... ){ // only pairs of interacting variables, with i < j
	DoubleImage M = ...;  // Put in M the interaction matrix between the 2 variables you picked, with x_i on the left side of the matrix and x_j on the top, and i < j. Be careful !
	
	Interact.push_back( make_pair(M, make_pair(i,j)) ); // convention : indices of variables are in increasing order, ie i<j => M = Mij and not Mji  (this is *IMPORTANT*)
	Neighbors[i].push_back( make_pair(j, nb_mat_inter) );
	Neighbors[j].push_back( make_pair(i, nb_mat_inter) );
	nb_mat_inter++;
      }

   Energy E0(Pot, Neighbors, Interact, 0); // Initial energy

  // Energy is ready. Needs canonization, identification, and minimization. Requires the energy to be in F0.
  // Skip next section (random energy)
  *****/


  /******************************************************************************/
  /********************* Code to create a random energy instead *****************/
  /******************************************************************************/

  /********* Parameters to create random energies automatically ********/

  int nb_var = 2 + ( rand() % 1000 ) ; // nb variables x_i  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Number of variables
  if (verbose_basic)
    cout << "Nb variables : " << nb_var << endl;
  vector< int > NbLabl(nb_var); // nb possible labels per variable x_i   // not very useful, since it is  Pot[i].size()

  for (vector<int>::iterator i = NbLabl.begin(); i != NbLabl.end(); ++i)  // random number of labels, not too big
    (*i) = int( pow(2, 1 + (rand() % 3)  ) + 0.1);   // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Put here (or later...) the number of labels of variable x_i
  
  int mean_neighborhood_size = 4; // to be multiplied by 2 to obtain the average number of neighbors of one variable // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Useless for your own energy


  /******** Creation of potentials, matrices and permutations ********/

  vector< vector<double> > Pot(nb_var); // potentials of each label for each variable
  for (int i = 0; i<nb_var; i++){
    //      NbLabl[i] = 8; // to fix a particular number of labels
    //      Pot[i] = vector<double>(NbLabl[i], 0); // to define a null potential
    // Pot[i] = random_vector(NbLabl[i]); // random potential but probably not admissible
    Pot[i] = generate_possible_potential(NbLabl[i]);  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Put here the individual potential of variable x_i, as a vector<double>
    if (verbose_a_lot)
      cout << "Pot x_" << i << " : " << Pot[i] << endl;
  }
    
  vector< vector< pair<int, int> > > Neighbors(nb_var); // list, for each variable, of the indices of its neighbors (with the index of the associated interaction matrix), initialized with "no neighbor"
  //  typedef vector< pair<DoubleImage, pair<int,int> > >   MIV; // to remember the type used in what follows
  Energy::MIV Interact; // Interaction matrices with the numberings of the variables involved
  int nb_mat_inter = 0;
  int aux_rand = (nb_var-1)/(2*mean_neighborhood_size);

  // one random permutation per variable x_i
  vector< vector<int> > permix(nb_var);
  for (int i = 0; i<nb_var; i++){  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  This paragraph shuffles randomly a random energy... remove it !
    permix[i] = generate_permutation(NbLabl[i]);
    // permix[i] = identity_permutation(NbLabl[i]); // identity permutation (1,2,3...)
    if (verbose_a_lot)
      cout << "Perm x_" << i << " : " << permix[i] << endl;
    Pot[i] = permute(Pot[i], permix[i]);
    if (verbose_a_lot)
      cout << "Pot' x_" << i << " : " << Pot[i] << endl;
  }

  for (int i = 0; i<nb_var; i++)
    for (int j = i+1; j<nb_var; j++)
      if ((aux_rand < 2) || (rand()%(aux_rand) == 0)){ // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Remove this condition that picked randomly 2 variables and choose only pairs of interacting variables, with i < j
	DoubleImage M;  // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  Put in M the interaction matrix between the 2 variables you picked, with x_i on the left side of the matrix and x_j on the top, and i < j. Be careful !
	//M = random_matrix(NbLabl[i], NbLabl[j]); // random matrix, probably not admissible
	M = generate_possible_matrix( int( log(NbLabl[i]) /log(2) + 0.1), int( log(NbLabl[j]) / log(2) + 0.1) , false  , true);
	if (verbose_a_lot){ cout << "M_" << i << "," << j << ":" << endl << M; } 
	M = permute_x(M, permix[i]); // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Remove these two lines, which suffle the random matrix
	M = permute_y(M, permix[j]);
	
	Interact.push_back( make_pair(M, make_pair(i,j)) ); // convention : indices of variables are in increasing order, ie i<j => M = Mij and not Mji  (this is *IMPORTANT*)
	Neighbors[i].push_back( make_pair(j, nb_mat_inter) );
	Neighbors[j].push_back( make_pair(i, nb_mat_inter) );
	nb_mat_inter++;
      }
  if (verbose_basic)
    cout << "Nb interactions : " << nb_mat_inter << endl;
  permix.clear();
  NbLabl.clear();




  /*****************************************/
  /************* Energy is ready ***********/
  /*****************************************/

  Energy E0(Pot, Neighbors, Interact, 0); // Initial energy
  Energy E; E.copy(E0); // Energy to be canonized (a copy)

    //  vector<double> Vx(n, 0); 
    //  vector<double> Vy(m, 0);
    //    canonise(M, Vx, Vy);
    //  double constante = canonise(Vx);
    //  constante += canonise(Vy);

  

  /*********************************************************************************************************/
  /*************************   canonization and familial representation  ***********************************/
  /*********************************************************************************************************/

  if (verbose_basic)
    cout << "Canonisation" << endl;

  canonise(E);

  //    cout << "Energy E0 : " << endl;
  //    E0.display_text();
  
  //    cout << endl << endl << "Energy E : " << endl;
  //    E.display_text();
  
  

  /******* test whether the canonisation changed the energy ******/
  if (test_canonisation){
    if (verbose_basic)
      cout << "Test canonisation correct ?..." << endl;
    
    if (! test_random_equality_energies(E0, E))
      cout << endl << BoldRed << "Canonisation failed..." << TextNormal << endl;
  }
    
  E0.empty();


  /****** find alphas and possible permutations ******/

  PreGraphMulti PG;
  construct_pregraph(PG, E, verbose_basic, verbose_summaries, verbose_multi, verbose_coherence, verbose_coherence_pot, verbose_global, cout);
  
  // Du blabla
  if (verbose_basic)
    cout << "Nb solutions (to the energy representation problem, ie finding permutations) [1 max now...] : " << PG.Solutions.size();
  int i_sol = 0;
  for (list< vector<int> >::iterator i = PG.Solutions.begin(); i != PG.Solutions.end(); ++i){
    i_sol++;
    if (verbose_summaries)
      cout << endl <<  "- Solution " << i_sol << " :" <<  (*i) << endl; 
    if ( (find_value((*i), -1)) != -1) 
      std::cout << endl <<  "- Solution " << i_sol << " :" <<  (*i) << endl; 
    
  }
  
  if (verbose_summaries)
    switch (PG.Solutions.size()) {
    case 0 : // red font
      std::cout << BoldRed << PG.Solutions.size() << ";" << TextNormal; break;
    case 1:
      std::cout << ";"; break;
    default:
      std::cout << BoldGreen << PG.Solutions.size() << ";" << TextNormal; 
    }
  std::cout.flush();

  if (PG.Solutions.empty())
    return;

  if (PG.Solutions.front().size() == 0){
    std::cout << BoldRed << "Huh ?" << TextNormal << endl;
  }

  //  return;

  /*********************************************************************************************************/
  /***********************************   graph and minimization   ******************************************/
  /*********************************************************************************************************/


  if (verbose_basic){ cout << endl;
    cout << TextGreen << "***********************************************************" << endl;
    cout << "*********" << TextRed << "  Creating the graph and minimization  " << TextGreen << "***********" << endl;
    cout << "***********************************************************" << TextNormal << endl;
  }


  vector<int> Resultat = minimize_familial_MRF(E, PG, true, verbose_graph);

  if (verbose_basic){ cout << "Optimal labelling:" << endl;
    cout << Resultat << endl;
  }


  // Cleanings
  PG.empty();
  E.empty();
  Resultat.clear();

  //    for (int k = 0; k<int( AlpesL.size() ); k++)
  //      for (int i = 0; i<int(AlpesL[k].size()); i++)
  //	AlpesL[k][i].empty();
  //    AlpesL.clear();
  
  //    Coherence.clear();
  //    Coeff_Pot.clear();
}


// permute labels of an energy
void permute(Energy& E, const vector<int>& ChoixPermut, const vector< vector<int> >& Perm){
  for (int i = 0; i<E.nb_var; i++)
    E.Pot[i] = permute(E.Pot[i], Perm[ ChoixPermut[i] ]);


  for (int m = 0; m<int(E.Interact.size()); m++){
    DoubleImage& M = E.Interact[m].first;
    M = permute_x(M,   Perm[   ChoixPermut[ E.Interact[m].second.first  ]   ]  );
    M = permute_y(M,   Perm[   ChoixPermut[ E.Interact[m].second.second ]   ]  );
  }
}

// opposite of the one above : find the index of the labels
vector<int> permute_inv_labl(const vector<int>& T, const vector<int>& ChoixPermut, const vector< vector<int> >& MetaPerm){
  vector<int> R(T.size());
  assert(T.size() == ChoixPermut.size());
  for (int i = 0; i<int(R.size()); i++){ // i : variable x_i;  T[i] : label of x_i, permuted
    R[i] = MetaPerm[ ChoixPermut[i] ][ T[i] ] ;
  }

  return R;
}




} // End namespace


























#endif
